#pragma once

#include "GraphicsAPI.h"
#include "TxformInfo.h"
#include "../Td3dDrawable.h"


		//
		// Set and get attr/function for compound data type:
		//  	e.g. D3DXCOLOR
		//
#define CompoundObjAttr(_dataType_, _attrName_)									\
			protected:															\
					_dataType_	f##_attrName_##;								\
			public:																\
					virtual void Set##_attrName_##(const _dataType_ &na); 		\
					virtual _dataType_##& Get##_attrName_##();  				\

		// For each defined CompoundObjAttr, the following macro should
		// be placed in the .cpp file to implement the basic Set and Get
		// function for the attribute
#define ImplementCompoundObjAttr(_className_, _dataType_, _attrName_)			\
				void _className_##::Set##_attrName_##(const _dataType_ &na) {	\
						f##_attrName_## = na; }									\
				_dataType_##& ##_className_##::Get##_attrName_##() { 			\
						return f##_attrName_##; }								\

		//
		// Set and get attr/function for pointer data type:
		//  	e.g. Tcircle*
		//
#define PointerObjAttr(_dataType_, _attrName_)									\
			protected:															\
					##_dataType_##*	f##_attrName_##;							\
			public:																\
					virtual void Set##_attrName_##(_dataType_ *na); 			\
					virtual _dataType_##* Get##_attrName_##();  				\

		// For each defined CompoundObjAttr, the following macro should
		// be placed in the .cpp file to implement the basic Set and Get
		// function for the attribute
#define ImplementPointerObjAttr(_className_, _dataType_, _attrName_)			\
				void _className_##::Set##_attrName_##(_dataType_ *na) {			\
						f##_attrName_## = na; }									\
				_dataType_##* ##_className_##::Get##_attrName_##() { 			\
						return f##_attrName_##; }								\

			
		//
		// Set and get attr/function for simple data type:
		//  	e.g. D3DXCOLOR (this is a DWORD)
		//  	   or int,
		//  	   or float, etc.
		//
#define SimpleObjAttr(_dataType_, _attrName_)									\
			protected:															\
					_dataType_	f##_attrName_##;								\
			public:																\
					virtual void Set##_attrName_##(const _dataType_ na); 		\
					virtual _dataType_ Get##_attrName_##() const;   			\
			
#define ImplementSimpleObjAttr(_className_, _dataType_, _attrName_)				\
				void _className_##::Set##_attrName_##(const _dataType_ na) {	\
							f##_attrName_## = na; }								\
				_dataType_  _className_##::Get##_attrName_##() const 	{ 		\
							return f##_attrName_##; }							\
			

		//
		// Set and get attr/function for boolean data type:
		//  	e.g. bool fVisiblility
		//
#define BoolObjAttr(_attrName_)													\
			protected:															\
					bool	f##_attrName_##;									\
			public:																\
					virtual void Enable##_attrName_##(); 						\
					virtual void Disable##_attrName_##(); 						\
					virtual bool  _attrName_##IsEnabled() const;	  			\

#define ImplementBoolObjAttr(_className_, _attrName_)							\
				void _className_##::Enable##_attrName_##() {					\
							f##_attrName_## = true; }							\
				void _className_##::Disable##_attrName_##() {					\
							f##_attrName_## = false; }							\
				bool  _className_##::_attrName_##IsEnabled() const 	{ 			\
							return f##_attrName_##; }							\
	
			

class Tcircle;

	// the origin for all graphics objects are located at
	// lower left-hand corner. Caller must ensure this is obeyed.
	//
class TgraphicsObject {

	private:
	protected:
		//
		// TxformInfo fXformInfo; // Transformation informaiton
		CompoundObjAttr(TxformInfo, XformInfo)

		//
		// D3DXVECTOR3	fVelocity; // speed
		CompoundObjAttr(D3DXVECTOR3, Velocity)

		//
		// Support for "collision circles" (OC stands for Object Space)
		PointerObjAttr(Tcircle, CollisionCirclesInOC)
		SimpleObjAttr(int, NumCollisionCircles)

		//
		// bool fVisibility; // is this object visible?
		BoolObjAttr(Visibility)
		
		//
		// bool fShowCollisionCircle // should we draw the collisionCircle?
		BoolObjAttr(ShowCollisionCircles)
	
	protected:
		//
		//
		// Following methods are for subclass to override
		// which will define/modify the detailed behavior
		// of each types of objects
		//
		//
		//
		virtual bool ShouldDrawObject() const;
					// should this object be drawn (e.g. is it visible?)

		virtual void SetObjectD3DAttributes(Td3dDrawable&, LPDIRECT3DDEVICE9);
					// subclass is responsible for setting the object's attributes
		
		
		virtual void UnSetObjectD3DAttributes(Td3dDrawable&, LPDIRECT3DDEVICE9);
					// subclass is responsible for setting the object's attributes
		
		virtual void DrawCollisionCircles(Td3dDrawable &drawable);
					// interesting for debuggin
					//

		virtual void DrawObjectGeom(Td3dDrawable&, LPDIRECT3DDEVICE9);
					// draws the object geometry
					// subclasses are responsible for
					// defining the implementaion of
					// this method.

	public:
		virtual ~TgraphicsObject();
		TgraphicsObject();

	
		void DrawObject(Td3dDrawable &drawable);
					// Draws the object

		virtual void UpdateLocation();
					//
					//  Updates location by fVelocity

		virtual void SetFirstPosition(const D3DXVECTOR3 &p);
		virtual void SetSecondPosition(const D3DXVECTOR3 &p);
		virtual const D3DXVECTOR3& GetFirstPosition() const;
					// First position is initial definition of the graphics object
					// second poisiont is often the "dragging out" of the object
					// Second position is "not standard" there is no way to access it.
					//

		virtual void MoveObjectTo(const D3DXVECTOR3 &d);
					// Moves object to this new location
	
		virtual void SetObjectD3DTransformation(ID3DXMatrixStack *stack) const; 
					// By default, simple calls the TxformInfo to set the matrix stack.
					// subclass may override this to implement scaling/rotation with non-origin
					// reference position.

		virtual void GetCollisionCirclesInWC(Td3dDrawable&, Tcircle *cc, int &n);
					// caller is responsible to make sure the incoming array is larger or equal to our internal
					// collision circle array. (find out with GetNumCollisionCircle()) function.
					// Caller is responsible for creating the circles and passing them in. After the caller is
					// done with the cc array, it is the caller's responsibility to manage the CC memory

		virtual bool WCcirclesTouchObject(Td3dDrawable &d, Tcircle*, int n, D3DXVECTOR3 &delta, float &distance);
					// decide if we collide with the circles in the incoming array
					// this collision decision is based on intersecting with the 
					// collision circles defined by the obj.
					//
					// 	delta - vector between us and the center of the obj->fCollisionCirle 
					// 	            that is causing the collsion
					//  distance - size of delta vector
					// 	
		
		virtual bool IntersectsBound(const T2dBound&);
					// Does the object intersect the bound? This method _MUST_ be override
					// by subclass (probably at leaf level). Always return false in the base class
					//

		//
		//
		// Following are virtual functions that describe behaviors
		// of drawable objects.  appropriate subclasses
		// should override these functions.
		//
		// the following set of behaviors allows TgraphicsObject understand
		// the language of all subclasses.  This is important for
		// "generic" graphics operation, e.g. selection in TobjectList
		//
		// 
		virtual void SetDrawColor(const D3DXCOLOR&);
		virtual D3DXCOLOR& GetDrawColor();

				//  point size
		virtual void SetPointSize(const float s);
		virtual float GetPointSize() const;

				// shadable attrbiute
		virtual void SetShadedMode(const WORD);
		virtual WORD GetShadedMode() const;

				// fillable attribute
		virtual void SetFillMode(const WORD);
		virtual WORD GetFillMode() const;
		virtual void SetTextureIndex(const int);
		virtual int  GetTextureIndex();
		virtual void EnableTexture();
		virtual void DisableTexture();
		virtual bool TextureIsEnabled();
		virtual void EnableAlphaBlending();
		virtual void DisableAlphaBlending();
		virtual bool AlphaBlendingIsEnabled();

	protected:
		void TgraphicsObject::XformCollisionCirclesToWC(D3DXMATRIX *m, Tcircle *c);
				// Transforms fCollisionCirclesInOC to c with the xformation matrix m

		bool CircleIntersectCircles(Tcircle &c, int num, Tcircle*, D3DXVECTOR3 &delta, float &distance);
				// utility function, computes if c (a Tcircle) intersects any of the input circles
				// if true:
				//
				// 	delta - vector between pt and the center of the closest circle
				//  distance - size of delta vector
				// 	
				//

		void CreateCollisionCircles(int n);
				// Creates n collision circles with proper color/fill/shade modes

};
