/////////////////////////////////////////////////////////////////////////////////////////////
//
//  This class is copied and modified from
//  
//  	Special Effects Game Programming with DIRECT X
//  	By Mason McCuskey
//  	Published by: Premier Press
//  	Date: 2002
//  	ISBN: 1-931841-06-3
// 
// Source code from Chapter 11: Fire (Ch11p2_TextureFire.cpp)
//
// By: Kelvin Sung
// Date: August 2002
//
//////////////////////////////////////////////////////////////////////

#include "TupwardEffect.h"



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
TupwardEffect::~TupwardEffect()
{
	if (fEffectBuffer1 != NULL) {
  		// delete upwards effect buffers
  		delete[] fEffectBuffer1; 
		fEffectBuffer1 = NULL;
	}

	if (fEffectBuffer1 != NULL) {
  		delete[] fEffectBuffer2; 
		fEffectBuffer2 = NULL;
	}

	if (fFuelPos != NULL)
		delete[] fFuelPos; 
}


bool TupwardEffect::CreateD3DTexture(
		int textureSize,
		int effectSize,
        int coolAmount
)
{
  	if (!fD3dDevice) { 
		EchoToStatusArea("DrawTexture creation failed: D3dDevice is not here!");
		return FALSE;
	}
  	if (textureSize < 2 || textureSize > 512 /* || !IsPowerOf2(textureSize) */) {
   		CString msg;
		msg.Format("DrawTexture size (%d) is bad!", textureSize); 
		EchoToStatusArea(msg);
		return FALSE;
  	}

  	fCoolAmount = coolAmount;
  	fTextureSize = textureSize;
	fEffectSize = effectSize;

  	// allocate upwards effect buffers
  	fEffectBuffer1 = new unsigned char[textureSize*textureSize];
  	fEffectBuffer2 = new unsigned char[textureSize*textureSize];

  	// set the first upwards effect buffer active, 2nd one as scratch surface
  	fActiveBuffer = fEffectBuffer1;
  	fScratchBuffer = fEffectBuffer2;

  	// clear the buffers
  	memset(fEffectBuffer1, 0, textureSize*textureSize);
  	memset(fEffectBuffer2, 0, textureSize*textureSize);

	LoadEffectPalette();

  	// create upwards effect texture
  	if (FAILED(D3DXCreateTexture(fD3dDevice, fTextureSize, fTextureSize, 1, 0, 
    		D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &fTexture))) {
    	EchoToStatusArea("TupwardEffect::RestoreDeviceObjects: can't create texture!");
		return FALSE;
  	}

	return TRUE;
}

void TupwardEffect::SetFuelSource(
		const D3DXVECTOR3 *fuelPos,		// array of positions
		int		size
)
{
	if (fFuelPos != NULL)
		delete [] fFuelPos;
	
	fFuelPos = new D3DXVECTOR3[size];
	for (int i=0; i<size; i++) {
		fFuelPos[i] = fuelPos[i];
	}
	fFuelSize = size;
}

/****************************************************************************

 Process: this function processes our upwards effect.  It takes two input buffers,
 the upwards effect dimensions, and the cooling amount.  It calculates the new upwards effect
 values from effectfield1 and puts them into effectfield2.

 ****************************************************************************/
void TupwardEffect::Process()
{
	// loop through all the upwards effect values...
	for (int y=0; y < fTextureSize; y++) {
	  	for (int x=0; x < fTextureSize; x++) {
	    	// these will store, temporarily, the upwards effect values immediately to the
	    	// left, right, top, and bottom of this upwards effect value.
	    	unsigned char effectvalue_left, effectvalue_right, effectvalue_bottom, effectvalue_top;
	    	int finaleffectvalue;
		
	    	// x+-1 and y+-1 aren't always that simple; we must account for wrapping
	    	// around the horizontal edge (not vertical, however).
	    	// so we calculate x/y +- 1 and store them temporarily.
	    	int xplus1, xminus1, yplus1, yminus1;
	    	xplus1 = x+1; if (xplus1 >= fTextureSize) xplus1=0;
	    	xminus1= x-1; if (xminus1 < 0) xminus1 = fTextureSize-1;
	    	yplus1 = y+1; if (yplus1 >= fTextureSize) yplus1=fTextureSize-1;
	    	yminus1= y-1; if (yminus1 < 0) yminus1 = 0;
	    
	    	// now we can get the upwards effect values of the neighboring pixels
	    	effectvalue_right = fActiveBuffer[(y*fTextureSize)+xplus1];
	    	effectvalue_left  = fActiveBuffer[(y*fTextureSize)+xminus1];
	    	effectvalue_bottom= fActiveBuffer[((yplus1)*fTextureSize)+x];
	    	effectvalue_top   = fActiveBuffer[((yminus1)*fTextureSize)+x];
	
	    	// now, the most important part-  calculate the new upwards effect value..
	    	finaleffectvalue = (effectvalue_left+effectvalue_right+effectvalue_top+effectvalue_bottom)/4;
	
	    	// subtract a certain amount to simulate the upwards effect "cooling."
	    	// this is where you'd apply your cooling map.
	    	finaleffectvalue -= fCoolAmount;
	
	    	// make sure that the subtraction of the coolamount didn't take us
	    	// below zero.
	    	if (finaleffectvalue < 0) finaleffectvalue = 0;
	
	    	// store the upwards effect value on the scratch array, up one line from where
	    	// it originally was.  This simulates the effects rising.
	    	fScratchBuffer[((yminus1)*fTextureSize)+x] = finaleffectvalue;
			fScratchBuffer[((y)*fTextureSize)+x] = finaleffectvalue;
	  	}
	}
	
  	// add fuel to the upwards effect.  This particular fueling method is one of my
  	// favorite; it creates little jet streams of effect.
	AddFuelToEffect();
}

int TupwardEffect::ComputeFuel(int index)
{
	return fActiveBuffer[index] + int(RandomNumber(float(-fEffectSize), float(fEffectSize)));
}

void TupwardEffect::AddFuelToEffect()
{
	if (fFuelDuration > 0) {
		fFuelDuration--;
	  	// we work in blocks of 2x1...
	  	for (int i=0; i < fFuelSize; i++) {
	    	int x= int(fFuelPos[i].x);
	    	int y= int(fFuelPos[i].y);
	    	
			int index = (y*fTextureSize)+x;
	
	    	// determine whether this particular spot gets fuel added or taken
	    	// away from it, by adding a number between (-fEffectSize..fEffectSize)
	    	int fuel = ComputeFuel(index);
		
	    	// we must be between 0-255.
	    	if (fuel > 255) fuel = 255;
			if (fuel < 0) fuel = 0;	
			
		
	    	// apply the new fuel value to two adjacent pixels.  This helps reduce
	    	// the "dithering" effect that the upwards effect is prone to.
	    	fScratchBuffer[index] = (unsigned char)fuel;
			if (index < (fTextureSize*fTextureSize-1) )
	    		fScratchBuffer[index+1] = (unsigned char)fuel;
	  	}
	}
}

void TupwardEffect::UpdateTexture()
{
	if (fActiveBuffer == NULL)
		return; 

  	Process();
  	CopyToTexture();

  	// swap the buffers.
  	unsigned char *temp = fActiveBuffer;
  	fActiveBuffer = fScratchBuffer;
  	fScratchBuffer = temp;
}

bool TupwardEffect::CopyToTexture(void)
{
  	HRESULT hr;
  	// lock texture
	D3DLOCKED_RECT lockedrect;
	::ZeroMemory(&lockedrect, sizeof(lockedrect));
	
	if (FAILED(hr = fTexture->LockRect(0, &lockedrect, NULL, 0))) return(false);
	
	// our texture surface is now locked, and we can use the pitch to traverse it.
	unsigned char *pSurfBits = static_cast<unsigned char *>(lockedrect.pBits);
	
	int index=0;
	
	for (int y=0; y<fTextureSize; y++) {
	  	for (int x=0; x < fTextureSize; x++) {
	    	// the upwards effect value at this position determines the color of this texel
	    	pSurfBits[index++] = fEffectPalette[fActiveBuffer[(y*fTextureSize)+x]].peBlue; // blue
	    	pSurfBits[index++] = fEffectPalette[fActiveBuffer[(y*fTextureSize)+x]].peGreen; // green
	    	pSurfBits[index++] = fEffectPalette[fActiveBuffer[(y*fTextureSize)+x]].peRed; // red
	    	pSurfBits[index++] = fEffectPalette[fActiveBuffer[(y*fTextureSize)+x]].peFlags; // alpha
			
	  	}
	  	// next line
	  	index += lockedrect.Pitch - (fTextureSize*4);
	}

	// unlock texture surface
	if (FAILED(hr = fTexture->UnlockRect(0))) return(false);
	return(true);
}
