#include "StdAfx.h"
#include "Td3dDrawable.h"

#include "./GraphicsAPI/TmyVertex.h"
#include "GraphicsAPI/GraphicsAPI.h"

//
// Create a connection to a piece of drawable area for D3D specific interactions (output and input).
//
BOOL Td3dDrawable::InitD3D(
	LPDIRECT3D9		 pD3d		// this is like the connection to the abstract D3D pipeline
)
{
	if (fD3DDevice != NULL) {
		fD3DDevice->Release();
		fD3DDevice = NULL;
	}

	// Get the current desktop display mode, for the default graphics card. If you have more than
	// one graphics cards on your machine, you may need to do something different here.
    D3DDISPLAYMODE d3ddm;
    if( FAILED( pD3d->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) ) {
        MessageBox("D3d Creation Failed (2)!");
		return FALSE;
	}

    // Set up the structure used to create the D3dDevice. Most parameters are
    // zeroed out. We set Windowed to TRUE, since we want to do D3d in a
    // window, and then set the SwapEffect to "discard", which is the most
    // efficient method of presenting the back buffer to the display.  And 
    // we request a back buffer format that matches the current desktop display 
    // format.
    D3DPRESENT_PARAMETERS d3dpp; 
		// 1. Init the entire D3DPRESENT structure to 0
    	ZeroMemory( &d3dpp, sizeof(d3dpp) );
		
		// 2. We are creating for one window
    	d3dpp.Windowed = TRUE;

		// 3. We will discard the buffer content after being swapped back (most efficient)
    	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    	d3dpp.BackBufferFormat = d3ddm.Format;

		// 4. Enable Z-buffer stenciling to support 3D ..
 		// in the future, we will have to do more here, but for now, for 2D this is sufficient.
		// go check out the D3D on-line manual to what other options are and what they mean.

    // Create the Direct3D device. Here we are using the default adapter (most
    // systems only have one, unless they have multiple graphics hardware cards
    // installed) and requesting the HAL (which is saying we want the hardware
    // device rather than a software one). Software vertex processing is 
    // specified since we know it will work on all cards. On cards that support 
    // hardware vertex processing, though, we would see a big performance gain 
    // by specifying hardware vertex processing.
	//
	// notice the somewhat mysterious "m_hWnd" parameter. This is the MFC-CWnd parameter
	// remember Td3dDrawable is a subclass from CWnd, so, we are using instance variable
	// from our parent class. m_hWnd is the "handler" to "Window" (abstract connection to
	// a window), think now we are creating a D3D-drawable area, that is connected to a MFC-Window
    if( FAILED( pD3d->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, m_hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &fD3DDevice ) ) )
    {
        MessageBox("D3d Creation Failed (3)!");
		return FALSE;
    }

	
	// Setup render states
	fD3DDevice->SetRenderState( D3DRS_LIGHTING,	FALSE);  
					// for 2D, we don't want to do lighting

	fD3DDevice->SetRenderState( D3DRS_SHADEMODE,	D3DSHADE_GOURAUD);
					// this says linearly interpolate color from vertex to vertex
					// instead of D3DSHADE_GOURAUD, later we will ...

	fD3DDevice->SetRenderState( D3DRS_ZENABLE,	D3DZB_FALSE);
					// z-buffer is for future 3D (451). We don't need this for 2D
					
	fD3DDevice->SetRenderState( D3DRS_CULLMODE,	D3DCULL_NONE);		
					// no culling for now. This does not make sense until we go to the third dimension (in 451)

	// increase point size, to help us see what we are drawing
	//
	float pointSize=2.0f;
	fD3DDevice->SetRenderState( D3DRS_POINTSIZE, *((DWORD*)&pointSize));
								// initialize to get big flat points, can be handy for debugging
		
	
	//******************** Vertex output format: to support drawing!!
	fD3DDevice->SetFVF(kMyVertexFormat);


	//
	// Now set up the system matrices
	//
	D3DXMATRIX identity;
	D3DXMatrixIdentity(&identity);
	//
	// when all three matrices are set to identity
	// D3D sees a normalized device space (-1 to +1), with origin
	// in the middle of the screen.
	// Now, we will initialize all these for now ...
	fD3DDevice->SetTransform(D3DTS_PROJECTION, &identity);
	fD3DDevice->SetTransform(D3DTS_VIEW, &identity);
	fD3DDevice->SetTransform(D3DTS_WORLD, &identity);

	//
	// Now create our matrix stack to help us transform from 
	// object space to world space
	if( FAILED( D3DXCreateMatrixStack(0 /* this flag is unsed */,
										&fObj2WorldStack) ) )
    {
        MessageBox("D3d Matrix Stack Creation Failed!");
		return FALSE;
    }
	
		
	if ( FAILED( fObj2WorldStack->LoadIdentity() ) ) 
	{
		MessageBox("D3d Matrix Stack Init to Identity failed!");
		return FALSE;

	}

	
	//
	// Now let's make sure we save our own window size
	CRect rect;
	this->GetClientRect(&rect);
	fWidth = rect.Width();
	fHeight = rect.Height();
	
	//
	// These two are Win32 function calls (from CWnd)
	//
	ShowWindow(SW_SHOWDEFAULT);
	UpdateWindow();

	// Initialize Graphics API's specific stuff
	InitializeGraphicsSystem(this);

	// Initilaize the file Texture library ...
	InitializeFileTextureLibrary();

    return TRUE;
}
